// src/completion.builder.ts
import { json } from "@synstack/json";
import { pipe } from "@synstack/resolved";
import { t } from "@synstack/text";
var CompletionBuilder = class _CompletionBuilder {
  _completion;
  constructor(completion2) {
    this._completion = completion2;
  }
  static get new() {
    return new _CompletionBuilder({});
  }
  static from(completion2) {
    return _CompletionBuilder.new.merge(completion2);
  }
  get $() {
    return this._completion;
  }
  $_(fn) {
    return pipe(this.$)._(fn);
  }
  async toSync() {
    return new _CompletionBuilder(await this.$);
  }
  /**
   * Override the values of the current builder with the provided values as object
   */
  merge(completion2) {
    return new _CompletionBuilder(
      pipe([this._completion, completion2])._(([old, added]) => ({
        ...old,
        ...added
      })).$
    );
  }
  // Todo: allow calling as a template directly
  /**
   * Set the system prompt
   */
  system(system) {
    return this.merge(
      pipe(system)._((system2) => ({
        system: system2
      })).$
    );
  }
  /**
   * Set the llm temperature
   * @argument temperature The temperature value between 0 and 1
   */
  temperature(temperature) {
    return this.merge(
      pipe(temperature)._((temperature2) => ({
        temperature: temperature2
      })).$
    );
  }
  /**
   * Set the maximum number of tokens to generate
   */
  maxTokens(maxTokens) {
    return this.merge(
      pipe(maxTokens)._((maxTokens2) => ({ maxTokens: maxTokens2 })).$
    );
  }
  /**
   * Set the stop sequences.
   *
   * If the LLM generates a stop sequence, the completion will stop.
   */
  stopSequences(sequences) {
    return this.merge(
      pipe(sequences)._((sequences2) => ({
        stopSequences: sequences2
      })).$
    );
  }
  /**
   * Set the top k tokens to consider for the completion
   */
  topK(topK) {
    return this.merge(
      pipe(topK)._((topK2) => ({
        topK: topK2
      })).$
    );
  }
  /**
   * Set the top p tokens to consider for the completion
   */
  topP(topP) {
    return this.merge(
      pipe(topP)._((topP2) => ({
        topP: topP2
      })).$
    );
  }
  /**
   * Reset the tools
   */
  clearTools() {
    return this.merge({
      toolsConfig: void 0
    });
  }
  /**
   * Set the tools available to the LLM
   *
   * @argument tools The tools to use
   * @argument requireToolUse Whether to require the LLM to use the tools
   */
  tools(tools, requireToolUse = false) {
    return this.merge(
      pipe(tools)._((tools2) => ({
        toolsConfig: {
          type: "multi",
          tools: tools2,
          requireToolUse
        }
      })).$
    );
  }
  /**
   * Set a single tool for the LLM to use. The LLM will always respond with this tool
   */
  tool(tool) {
    return this.merge(
      pipe(tool)._((tool2) => ({
        toolsConfig: {
          type: "single",
          tool: tool2
        }
      })).$
    );
  }
  /**
   * Set the stop reason for the completion
   */
  stopReason(stopReason) {
    return this.merge(
      pipe(stopReason)._((stopReason2) => ({
        stopReason: stopReason2
      })).$
    );
  }
  /**
   * Set the tokens usage
   */
  usage(usage) {
    this.merge(
      pipe(usage)._((usage2) => ({
        usage: usage2
      })).$
    );
  }
  /**
   * Add usage to existing usage
   */
  addUsage(usage) {
    return this.merge(
      pipe([this.$, usage])._(([$, usage2]) => ({
        usage: {
          inputTokens: $.usage?.inputTokens ?? 0 + usage2.inputTokens,
          outputTokens: $.usage?.outputTokens ?? 0 + usage2.outputTokens
        }
      })).$
    );
  }
  /**
   * Set messages
   */
  messages(messages) {
    return this.merge(
      pipe(messages)._((messages2) => ({
        messages: messages2
      })).$
    );
  }
  /**
   * Add messages
   */
  addMessages(messages) {
    return this.merge(
      pipe([this.$, messages])._(([$, messages2]) => ({
        messages: [...$.messages ?? [], ...messages2]
      })).$
    );
  }
  /**
   * Add assistant message
   */
  addAssistantMessage(message) {
    return this.merge(
      pipe([this.$, message])._(([$, message2]) => {
        const $messages = $.messages ?? [];
        const lastMessage = $.messages.at(-1);
        if (!lastMessage)
          return {
            messages: [message2]
          };
        if (lastMessage.role !== "assistant")
          return {
            messages: [...$messages, message2]
          };
        return {
          messages: [
            ...$messages.slice(0, -1),
            {
              ...lastMessage,
              content: [...lastMessage.content, ...message2.content]
            }
          ]
        };
      }).$
    );
  }
  /**
   * Execute completion on Llm runner
   */
  run(runner) {
    return runner.runChatCompletion(this);
  }
  lastMessage() {
    return pipe(this.$)._((completion2) => {
      const lastMessage = completion2.messages?.at(-1);
      if (!lastMessage) throw new EmptyMessagesException(completion2);
      return lastMessage;
    }).$;
  }
  /**
   * Get last message as text
   */
  lastMessageText() {
    return pipe(this.lastMessage())._(
      (message) => message.content.filter((c) => c.type === "text").map((c) => c.text).join("")
    ).$;
  }
  /**
   * Get last message tool calls
   */
  lastMessageToolCalls() {
    return pipe(this.lastMessage())._((message) => {
      if (message.role !== "assistant") return [];
      return message.content.filter((c) => c.type === "tool_call");
    }).$;
  }
};
var EmptyMessagesException = class extends Error {
  constructor(completion2) {
    super(t`
      No messages in current completion state:
        ${json.serialize(completion2, { pretty: true })} 
    `);
  }
};
var completion = CompletionBuilder.new;

// ../shared/src/ts.utils.ts
function never(neverCheck, value) {
  throw new Error(
    `Unexpected value, received: ${JSON.stringify(value ?? neverCheck, null, 2)}`
  );
}

// src/message.builder.ts
import { pipe as pipe2 } from "@synstack/resolved";
import { t as t2, tParse } from "@synstack/text";
var userMsg = (template, ...values) => pipe2(t2(template, ...values))._(tParse)._(
  (parts) => parts.map((v) => {
    if (typeof v === "string")
      return { type: "text", text: v };
    if (v.type === "base64")
      return {
        type: "image",
        image: v
      };
    never(v.type);
  })
)._(
  (content) => ({
    role: "user",
    content
  })
).$;
var assistantMsg = (template, ...values) => {
  return pipe2(t2(template, ...values))._(tParse)._(
    (parts) => parts.map((v) => {
      if (typeof v === "string")
        return { type: "text", text: v };
      if (v.type === "tool_call") return v;
      never(v.type);
    })
  )._(
    (content) => ({
      role: "assistant",
      content
    })
  ).$;
};
export {
  CompletionBuilder,
  assistantMsg,
  completion,
  userMsg
};
//# sourceMappingURL=llm.index.js.map