import { ZodSchema, AnyZodObject, z } from 'zod';

/**
 * Merges two objects
 */
type Merge<T, U> = {
    [K in keyof T | keyof U]: K extends keyof T ? T[K] : K extends keyof U ? U[K] : never;
};
/**
 * One or more values
 */
type OneToN<T> = [...T[], T];

type $Partial<T> = Partial<T>;
declare namespace Llm {
    type Tool<TName extends string = string, TSchema extends ZodSchema = AnyZodObject> = {
        name: TName;
        schema: TSchema;
    };
    type Completion = {
        temperature: number;
        maxTokens: number;
        messages: Array<Llm.Message>;
        system?: string;
        topK?: number;
        topP?: number;
        stopSequences?: string[];
        toolsConfig?: Llm.Completion.ToolConfig;
        usage?: Completion.Usage;
        stopReason?: Completion.StopReason;
    };
    namespace Completion {
        type StopReason = "end" | "max_tokens" | "tool_call" | "stop_sequence";
        type Usage = {
            inputTokens: number;
            outputTokens: number;
        };
        type ToolConfig = ToolConfig.Single | ToolConfig.Multi;
        namespace ToolConfig {
            type Type = "single" | "multi";
            type Single<TTool extends Llm.Tool = Llm.Tool> = {
                type: "single";
                tool: TTool;
            };
            type Multi<TTools extends Array<Llm.Tool> = Array<Llm.Tool>, TRequire extends boolean = boolean> = {
                type: "multi";
                tools: TTools;
                requireToolUse: TRequire;
            };
        }
        type WithResponse = Completion & {
            stopReason: StopReason;
            usage: Usage;
        };
        namespace Response {
            type Part = {
                stopReason: StopReason;
                usage: Usage;
            };
        }
        type Partial = $Partial<Completion>;
    }
    type Message = User.Message | Assistant.Message;
    namespace Message {
        type Role = User.Role | Assistant.Role;
        type Content = Content.Text | Content.Image | Content.ToolCall | Content.ToolResponse;
        namespace Content {
            type Text = {
                type: "text";
                text: string;
            };
            type Image = {
                type: "image";
                image: Image.Base64;
            };
            namespace Image {
                type Base64 = {
                    type: "base64";
                    data: string;
                    mimeType: string;
                };
            }
            type ToolCall<TTool extends Llm.Tool = Llm.Tool> = {
                type: "tool_call";
                toolCallId: string;
                toolName: TTool["name"];
                toolArgs: z.output<TTool["schema"]>;
            };
            namespace ToolCall {
                type ResponseFromToolConfig<TConfig extends Completion.ToolConfig> = TConfig extends Completion.ToolConfig.Single ? [ToolCall<TConfig["tool"]>] : TConfig extends Completion.ToolConfig.Multi ? TConfig["requireToolUse"] extends true ? OneToN<{
                    [K in keyof TConfig["tools"]]: TConfig["tools"][K] extends infer U ? U extends Tool ? ToolCall<U> : never : never;
                }[number]> : Array<{
                    [K in keyof TConfig["tools"]]: TConfig["tools"][K] extends infer U ? U extends Tool ? ToolCall<U> : never : never;
                }[number]> : Array<ToolCall>;
            }
            type ToolResponse = {
                type: "tool_response";
                toolCallId: string;
                toolOutput: any;
            };
        }
    }
    namespace User {
        type Role = "user";
        type Content = Message.Content.Text | Message.Content.Image | Message.Content.ToolResponse;
        type Message = {
            role: Role;
            content: Array<Content>;
        };
    }
    namespace Assistant {
        type Role = "assistant";
        type Content = Message.Content.Text | Message.Content.ToolCall;
        type Message = {
            role: Role;
            content: Array<Content>;
        };
    }
}

export { Llm as L, type Merge as M, type OneToN as O };
